
import ahocorasick

class ZiSegmenter:
    def __init__(self, root_words, prefixs=[], suffixs=[],max_split=3):
        self.max_split = max_split
        self.root_words = root_words
        self.prefixs = set(x for x in prefixs)
        self.suffixs = set(x for x in suffixs)
        self.rootAC = ahocorasick.Automaton()
        for i,x in enumerate(self.root_words):
            self.rootAC.add_word(x, x)
        self.rootAC.make_automaton()

    def token_root(self, word):
        matchs = list(self.rootAC.iter(word))
        if not matchs:
            return [word, None, None]
        length = max(len(x[1]) for x in matchs)
        long_match = [x for x in matchs if len(x[1]) == length]
        longest_match = long_match[(len(long_match)-1)//2]
        end,root = longest_match
        prefix = word[:end-len(root)+1]
        suffix = word[end+1:]
        return [prefix, root, suffix]

    def token_prefix(self, grams):
        tokens = []
        for i in range(min(self.max_split,len(grams))):
            if not grams:
                break
            for i in range(len(grams)):
                a = grams[:len(grams)-i]
                if a in self.prefixs:
                    tokens.append(a)
                    grams = grams[len(a):]
                    break
        return tokens

    def token_suffix(self, grams):
        tokens = []
        for i in range(min(self.max_split, len(grams))):
            if not grams:
                break
            for i in range(len(grams)):
                a = grams[i:]
                if a in self.suffixs:
                    tokens.insert(0, a)
                    grams = grams[:-len(a)]
                    break
        return tokens

    def token_word(self, word):
        [prefix, root, suffix] = self.token_root(word)
        if not root:
            return [prefix, root, suffix]
        heads = []
        tails = []
        if prefix:
            heads = self.token_prefix(prefix)
        if suffix:
            tails = self.token_suffix(suffix)
        return [heads, root, tails]


def test_segmenter():
    line = "'〇㎡[คุณจะจัดพิธีแต่งงานเมื่อไรคะัีิ์ื็ํึ]Ⅷpays-g[ran]d-blanc-élevé » (白高大夏國)😀'\x0000熵"
    roots = ['la', 'a', 'ay', 'le']
    prefixs = ['e', 'l']
    suffixs = ['n', 'e', 'v']
    # from ZiTokenizer.UnicodeTokenizer import UnicodeTokenizer
    from UnicodeTokenizer import UnicodeTokenizer

    words = UnicodeTokenizer().tokenize(line)
    cutter = ZiSegmenter(roots, prefixs, suffixs)
    for word in words:
        tokens = cutter.token_word(word)
        print(word, tokens)

if __name__ == "__main__":

    test_segmenter()


# -*- coding: utf-8 -*-
import os

from logzero import logger

from ZiTokenizer import He2Zi
from ZiTokenizer.He2Zi import JieGou, GouJian


def slim(v):
    if len(v) <= 3:
        return v
    for i in range(1, len(v)-1):
        if v[i] not in JieGou:
            w = v[0]+v[i]+v[-1]
            return w
    assert len(v) == 3
    return v


def loadHeZi(path, JiZi, max_len=50):
    doc = open(path).read().splitlines()
    doc = [x.split('\t') for x in doc]
    doc = [(k, v) for k, v in doc if 0 < len(v) <= max_len and not set(v) - JiZi]
    HeZi = {k: v for k, v in doc}
    slim2Zi = {v: k for k, v in HeZi.items()}
    values = ''.join(HeZi.values())
    values = list(set(values))
    values.sort()
    jieGou = ''.join(set([x[0] for x in slim2Zi]))
    max_len = max(len(x) for x in slim2Zi)
    logger.info(
        f"  {path} {len(doc)}  JiZi:{len(JiZi)} --> loadHeZi {len(HeZi)}  values:{len(values)} slim2Zi:{len(slim2Zi)} jieGou:{jieGou} max_len:{max_len}")
    return HeZi, values, slim2Zi, jieGou, max_len


class ZiCutter:
    def __init__(self, dir=None, do_lower_case=True, max_len=50, k=10):
        """
        """
        self.do_lower_case = do_lower_case
        self.max_len = max_len
        self.k = k
        self.here = os.path.dirname(__file__)
        self.HanZiDir = os.path.join(self.here, "HanZi")
        self.dir = dir
        if dir and os.path.exists(dir):
            HeZiPath = os.path.join(dir, "HeZi.txt")
            if os.path.exists(HeZiPath):
                self.load(dir)

    def load(self, dir):
        HeZiPath = os.path.join(dir, "HeZi.txt")
        JiZiPath = os.path.join(dir, "JiZi.txt")
        if os.path.exists(JiZiPath):
            JiZi = open(JiZiPath).read().splitlines()
        else:
            JiZi = GouJian
        JiZi = set(JiZi)
        logger.info(f"{JiZiPath} load  JiZi:{len(JiZi)}")

        HeZi, values, slim2Zi, jieGou, max_len = loadHeZi(HeZiPath, JiZi, self.max_len)
        self.slim2Zi = slim2Zi
        self.HeZi = HeZi
        self.jieGou = jieGou
        self.max_len = max_len
        self.vocab = values
        logger.info(f"{dir} loaded vocab:{len(self.vocab)}")

    def build(self, folder, roots, max_len=10):
        logger.warning(f" {folder} building")
        vocab = set(GouJian) | set(x for x in roots)
        JiZi = [x for x in vocab if len(x) == 1]
        logger.info(f"receive roots:{len(roots)} JiZi:{len(JiZi)}")

        ChaiZiPath = os.path.join(self.HanZiDir, "ChaiZi.txt")
        YiTiZiPath = os.path.join(self.HanZiDir, "YiTiZi.txt")
        HeZiPath = os.path.join(folder, "HeZi.txt")
        JiZiPath = os.path.join(folder, "JiZi.txt")
        He2Zi.build(JiZi, ChaiZiPath, YiTiZiPath, HeZiPath, JiZiPath, max_len)
        self.load(folder)

    def cutHanzi(self, zi) -> str:
        ids = self.HeZi.get(zi, zi)
        return ids

    def tokenize(self, line):
        tokens = [self.cutHanzi(x) for x in line]
        return tokens

    def combineHanzi(self, chars):
        tokens = []
        i = 0
        while i < len(chars):
            ok = False
            x = chars[i]
            if x in self.jieGou:
                for j in range(2, self.max_len):
                    D = ''.join(chars[i:i+j])
                    if D in self.slim2Zi:
                        tokens.append(self.slim2Zi[D])
                        i += len(D)
                        ok = True
                        break

            if not ok:
                tokens.append(x)
                i += 1
        return tokens



# -*- coding: utf-8 -*-

import unicodedata


class UnicodeTokenizer:
    def __init__(self,  do_lower_case=True, never_split=[], highUnicodePoint=10000, split_digit=False):
        self.do_lower_case = do_lower_case
        self.highUnicodePoint = highUnicodePoint
        self.split_digit = split_digit
        self.never_split = set(x for x in never_split)

    def split_blank(self, line):
        tokens = [x.strip() for x in line.split() if x.strip()]
        return tokens

    def split_marks(self, line, marks):
        tokens = []
        for i, x in enumerate(line):
            if i == 0:
                tokens.append(x)
            elif marks[i] or marks[i-1]:
                tokens.append(x)
            else:
                tokens[-1] += x
        return tokens

    def normalize(self, line,  normal_type="NFD"):
        l = unicodedata.normalize(normal_type, line)
        return l

    def split_highUnicodePoint(self, line):
        marks = [ord(x) >= self.highUnicodePoint for x in line]
        return self.split_marks(line, marks)

    def split_category(self, line):
        if len(line) == 1:
            return [line]
        elif len(line) == 0:
            return []
        categorys = [unicodedata.category(x) for x in line]
        names = [unicodedata.name(x).split()[0] if categorys[i][0] in 'LN' else None for i, x in enumerate(line)]
        tokens = []
        for i, x in enumerate(line):
            if i == 0:
                tokens.append(x)
            elif categorys[i][0] == categorys[i-1][0] == 'L':
                if names[i] == names[i-1]:
                    tokens[-1] += x
                else:
                    tokens.append(x)
            elif not self.split_digit and categorys[i][0] == categorys[i-1][0] == 'N':
                if names[i] == names[i-1]:
                    tokens[-1] += x
                else:
                    tokens.append(x)
            else:
                if categorys[i] != 'Mn':
                    tokens.append(x)
                if categorys[i-1] == 'Mn':
                    tokens.append('')

        tokens = [x.strip() for x in tokens if x.strip()]
        return tokens

    def split_word(self, x):
        tokens = []
        if self.do_lower_case:
            x = self.normalize(x.lower())
        us = self.split_blank(x)
        for u in us:
            vs = self.split_highUnicodePoint(u)
            for v in vs:
                w = self.split_category(v)
                tokens += w
        return tokens

    def tokenize(self, line):
        words = self.split_blank(line)
        tokens = []
        for x in words:
            if x in self.never_split:
                tokens.append(x)
            else:
                tokens += self.split_word(x)
        return tokens

    def detokenize(self, tokens):
        l = tokens[0]
        for i, x in enumerate(tokens):
            if i == 0:
                continue
            a = tokens[i-1][-1]
            b = x[0]
            if max(ord(a), ord(b)) < self.highUnicodePoint and unicodedata.category(a)[0] == unicodedata.category(b)[0] and unicodedata.category(a)[0] in 'LN':
                l += ' '
            l += x
        return l


def test_UnicodeTokenizer():
    doc = ["'〇㎡[คุณจะจัดพิธีแต่งงานเมื่อไรคะัีิ์ื็ํึ]Ⅷpays-g[ran]d-blanc-élevé » (白高大夏國)😀熇'\x0000𧭏2022２０１９\U0010ffff",
           "대한민국의Ⅷ首先8.88设置 st。art_new_word=True 和 output=[açaí]，output 就是最终 no such name",
           "的输出คุณจะจัดพิธีแต่งงานเมื่อไรคะ탑승 수속해야pneumonoultramicroscopicsilicovolcanoconiosis",
           "하는데 카운터가 어디에 있어요ꆃꎭꆈꌠꊨꏦꏲꅉꆅꉚꅉꋍꂷꂶꌠلأحياء تمارين تتطلب من [MASK] [PAD] [CLS][SEP]",
           '''est 𗴂𗹭𘜶𗴲𗂧, ou "phiow-bjij-lhjij-lhjij", ce que l'on peut traduire par « pays-grand-blanc-élevé » (白高大夏國).''',
           'วรรณพงษ์เป็นนักศึกษาชั้นปีที่หนึ่ง เรียนสาขาวิทยาการคอมพิวเตอร์และสารสนเทศคณะวิทยาศาสตร์ประยุกต์และวิศวกรรมศาสตร์อยู่ที่มหาวิทยาลัยขอนแก่นวิทยาเขตหนองคายยืมคืนทรัพยากรห้องสมุดเอกสารสัมมนาคอมพิวเตอร์ปัญญาประดิษฐ์กับการพัฒนาเกมแมวกินปลาหิวววไหมหลักสูตรใหม่สดสดทนได้',
           'ສົມເດັດພະເຈົ້າຢູ່ຫົວບໍຣົມໂກດຊົງທຳນຸບຳລຸງບ້ານເມືອງແລະພະສາດສະໜາຈົນກ່າວໄດ້ວ່າກຸງສີອະຍຸທະຢາໃນສະໄໝພະອົງນັ້ນເປັນຍຸກທີ່ບ້ານເມືອງດີ ມີຂຸນນາງຄົນສຳຄັນທີ່ເຕີບໂຕໃນເວລາຕໍ່ມາ ໃນລາຊະການຂອງພະອົງຫຼາຍຄົນ ເຊັ່ນ ສົມເດັດພະເຈົ້າກຸງທົນບຸລີ, ພະບາດສົມເດັດພະພຸດທະຍອດຟ້າຈຸລາໂລກມະຫາລາດ ເປັນຕົ້ນ ໃນທາງດ້ານວັນນະຄະດີກໍມີກະວີຄົນສຳຄັນ ເຊັ່ນ ເຈົ້າຟ້າທຳມາທິເບດໄຊຍະເຊດສຸລິຍະວົງ ກົມມະຂຸນເສນາພິທັກ ຫຼືເຈົ້າຟ້າກຸ້ງ ເຊິ່ງເປັນພະໂອລົດ ເປັນຕົ້ນ'
           ]
    unicodeTokenizer = UnicodeTokenizer()
    for line in doc:
        tokens = unicodeTokenizer.tokenize(line)
        print(tokens)


if __name__ == "__main__":
    from logzero import logger

    line = "대한민국의'〇㎡[คุณจะจัดพิธีแต่งงานเมื่อไรคะัีิ์ื็ํึ]Ⅷpays-g[ran]d-blanc-élevé » (白高大夏國)😀熇'\x0000𧭏2022２０１９\U0010ffff"
    # line = "art_new_word=True"
    tokenizer = UnicodeTokenizer()
    logger.info((tokenizer.split_blank(line)))
    # line = "=True"

    tokenizer = UnicodeTokenizer(split_digit=True)
    tokens = tokenizer.tokenize(line)
    logger.info(tokens)
    logger.info(tokenizer.detokenize(tokens))
    import timeit
    # re=timeit.timeit("''.join(chr(x) for x in range(int(1e6))) ")
    # logger.info(re)

    import time
    t0 = time.time()
    for i in range(10000):
        # chr(i)  # ValueError: chr() arg not in range(0x110000)
        tokenizer.tokenize(line)
    t1 = time.time()
    logger.info(t1-t0)

    test_UnicodeTokenizer()

"""
 '〇㎡[ค ณ จะจ ด พ ธ แ ต ง งานเม อ ไรคะ]ⅷpays-g[ran]d-blanc-e l eve»(白高大夏國)😀熇'00𧭏２０１９􏿿
"""



# -*- coding: utf-8 -*-

from logzero import logger
import copy
star = '䖵'

"""
    𤍽	𤑔 k,v  异体字\t本体字
    HeZi[𤑔]=HeZi[𤍽] if 𤍽 in 𤑔
    HeZi[v]=HeZi[k] if k in v
异体字 冃	帽
    """

JieGou = "〾⿰⿱⿲⿳⿴⿵⿶⿷⿸⿹⿺⿻"
# 591
GouJian = "⿰⿱⿲⿳⿴⿵⿶⿷⿸⿹⿺⿻⿼⿽⿾⿿〇〾㇀㇁㇂㇃㇄㇅㇆㇇㇈㇉㇊㇋㇌㇍㇎㇏㇐㇑㇒㇓㇔㇕㇖㇗㇘㇙㇚㇛㇜㇝㇞㇟㇠㇡㇢㇣㇤㇥㇦㇧㇨㇩㇪㇫㇬㇭㇮㇯㐄㐅㐫㐱㒳㒼㓁㔾㕯㠯㡀㢴㣇㦰㫃㲋㸦䇂䒑䶹一丁丂七丄丅丆万丈三上下丌与丏丑丘丙业丣严丨丩丬中丮丯丰丱丵丶丷丿乀乁乂乇乑乙乚乛九习乡亅二亍亏五井亜亞亠亥产亯人亻亼亾仌从來侖儿兀兂兆先克入八公六冂冃冄冈冉冋册冎冏冖冘冫几凵刀刂刃刄刅力勹匕北匚匸十卂千卄卅午卌南卜卝卤卩卯厂厃厄厶厷叀又叕叚口史吅咅咼啇喿嘼囗四囟囧囪囱土圥坴堇士壬壴夂夅夆夊夋夌夕夗大夨天夫夬女娄婁子宀寅寸尃小尚尞尢尣尸尺屮屰山屵巛巜川州巠巤工巨己巳巴巾巿帀干幵并幷幺广庚廌廴廾廿开弋弓彐彑彖彡彳心忄戈戉戊戌戍我戶户戼手扌支攴攵文斗斤方无旡日昜昷曰曷月木朩未本朮朱朿東枼桼欠止步歹歺殳毋毌母比毚毛氏氐民气水氵氶氺火灬爪爫爭父爻爾爿片牙牛牜犬犭犮玄玉王瓜瓦甘生用田甲申甶甾畀畐疋疌疒癶癸白皀皋皮皿盍目睘矛矢石示礻禸禺禾穴立竹米粦糸糹絲纟缶网罒罓羊羽翏老耂而耒耳聿肀肉肙臣自至臼臽與舌舛舟艮色艸艹菐萬虍虎虫血行衣衤襾西覀覃見见角言訁讠谷豆豕豸貝贝赤走足身車车辛辰辵辶邑酉釆采里金釒钅長镸长門门阜阝隶隹隺雚雨霝靑青非面靣革韋韦韭音頁页風风飛飞食飠饣首香馬马骨高髟鬥鬯鬲鬼魚鱼鳥鸟鹵鹿麥麦麻黃黄黍黑黹黽黾鼎鼓鼠鼻齊齐齒齿龍龙龜龠龰龴龵龶龷龸龹龻爫艹𠀁𠂊𠂤𠂭𠃑𠃬𠆢𠕁𠘨𠤎𠫓𡭔𡿨𣎳𣶒𤴓𤴔𦈢𦍌𦘒𦣝𦣞𦥑𧾷𨸏𩙿"


def chai(JiZi: set, ChaiZi: list, YiTi: dict, max_len=5, n_epoch=5):
    HeZi0 = {}
    for k, v in ChaiZi:
        if k in JiZi:
            HeZi0[k] = k
        elif k in YiTi and YiTi[k] in JiZi:
            HeZi0[k] = '〾'+YiTi[k]
        else:
            HeZi0[k] = v

    dic0 = copy.deepcopy(HeZi0)
    for epoch in range(n_epoch):
        dic1 = {}
        for k, v in dic0.items():
            if k == v:
                continue
            if ord(k) < 10000:
                continue
            if len(v) > max_len:
                continue
            if not set(v) - JiZi:
                dic1[k] = v
            else:
                u = ''.join(dic0.get(x, x) for x in v)
                if len(u) > len(v):
                    dic1[k] = u

        base0 = set(''.join(x for x in dic0.values()))
        base1 = set(''.join(x for x in dic1.values()))
        logger.info((f"epoch:{epoch} base:{len(base0)} --> {len(base1)} "))
        dic0 = dic1

    HeZi = {k: v for k, v in dic0.items() if not set(v) - JiZi}
    Base = set(''.join(HeZi.values()))

    diff = Base-JiZi
    if diff:
        logger.error(f"Base-JiZi:{len(diff)}  {''.join(diff)[:1000]}")
    assert len(diff) == 0
    giveup = HeZi0.keys()-HeZi.keys()
    logger.info(f" chars:{len(HeZi)} -> {len(HeZi)}  giveup v:{len(giveup)} {''.join(giveup)[:1000]}")
    logger.info(f" jizi:{len(JiZi)} -> {len(Base)} useless k:{len(JiZi-Base)} {''.join(JiZi-Base)[:1000]}")

    return HeZi, Base


def build(JiZi, ChaiZiPath, YiTiZiPath,  HeZiPath, JiZiPath, max_len=50):
    JiZi = [x for x in JiZi if x]
    JiZi = set(JiZi)

    doc = open(YiTiZiPath).read().splitlines()
    YiTiZi = [x.split('\t') for x in doc]
    YiTiZi = {k: v for k, v in YiTiZi}

    doc = open(ChaiZiPath).read().splitlines()
    ChaiZi = [x.split('\t') for x in doc]

    logger.info(f"JiZi:{len(JiZi)} ChaiZi:{len(ChaiZi)} YiTiZi:{len(YiTiZi)}")
    HeZi, Base = chai(JiZi, ChaiZi, YiTiZi, max_len=max_len)

    Base = list(Base)
    Base.sort()
    with open(JiZiPath, "w") as f:
        for x in Base:
            f.write(x+'\n')

    chars = list(HeZi)
    chars.sort()
    with open(HeZiPath, "w") as f:
        for x in chars:
            l = f"{x}\t{HeZi[x]}"
            f.write(l+'\n')

    logger.info(f"HeZi build success -> {HeZiPath}  {JiZiPath}")


if __name__ == "__main__":
    JiZi = GouJian
    build(JiZi, ChaiZiPath="ZiTokenizer/HanZi/ChaiZi.txt", YiTiZiPath="ZiTokenizer/HanZi/YiTiZi.txt",
          HeZiPath="demo/He2Ji.txt", JiZiPath="demo/JiZi.txt")


"""
[I 230419 20:24:39 He2Zi:76] JiZi:591 ChaiZi:96935 YiTiZi:24237
[I 230419 20:24:40 He2Zi:48] epoch:0 base:11130 --> 2339 
[I 230419 20:24:40 He2Zi:48] epoch:1 base:2339 --> 707 
[I 230419 20:24:40 He2Zi:48] epoch:2 base:707 --> 555 
[I 230419 20:24:40 He2Zi:48] epoch:3 base:555 --> 548 
[I 230419 20:24:40 He2Zi:48] epoch:4 base:548 --> 548 
[I 230419 20:24:41 He2Zi:59]  chars:87277 -> 87277  giveup v:9658 蔖𪻨𧻾㷪𤧛鮎𦙃𫦡琭乩斵檜𣋘𨒺𭳋𲈜𫪨崡𱟸馦䔷𮦏豠𪻮𰽏玮吅𥺇䏅𣵛𠰷𫴫䅤𪐵𡏄𠈇𡂬𢶧�𬒜𦦱臇臇𢐭𣜖𤨩𭸮𥲍𡇹
𩩏𡖣𠝎𬁃𢊯醩非𤫨𡰺𦅂䐌𬿄𤴗𡊉𤌮𨽅瑞𤽆㛯𣼁𣎹𦮊𪘲䓬𦋙絲𨾗𢧗㾄𬱗𡭿𪼱𠛤𫾇𤪶𡷭𩷘𡳧𬍷𦄓𮄗𣇺𥡍𰩈𧗎𦤩喳璒㖂𠪊𫞧𥢥垛𬟽词𰔁箄𭻐𱠴𤆣𫁝𭶆𬽙𠔍褄𲃥𱞋㯅𫅮𤺉𣥊𮏭𩾅𦺽𰢏𤣴𨸚䰗𬕼𱷗𫣂𠶵
㺱鵪皈𣼜𩃓𡷢𣭾楂𥼊𰩲𰸭婰𨯤𮐼𨖠𪼍俺乡溈𥙷烫𭪭𭔥䶵𰓡䨡𨐁靜𥃙𠛲诅𩓃𬾥𩌓鞞𥇗𠙐𫽖柂𲀚𦋚𰇯𠣉巨㻊𪜑𧻷牍𰩐𭘜𤩘望𱬥𤩎䄋𱷈𠮤𥐨𥔝�𥠅𢧰𰷬栀栀𩉐䑮𪃶娩唵𰗗𲇷𪼟𬴯𤪾𥅤敮𤱌𰢇𱒲𣆼绣 
𱎁琨陠𱋵馳𱤋蒹𭾫𢻫珽𱤩㥁戶𥩯𣀊𬌒𰈱𥝽瑝𤦯𨋜𩆈𠓋𥉼𣼇𫒜𡟤㯛𠧉𤥻𡠝𦈆酦𱔉㻁𧒢𩜏㕀姐䁆𪴕𰢀𬴴𣹋𡮡㿠𢠿𥽠𤴋𢯂𠢣𫩞琍焄𪽿𠶦𭴙𮐖�瓙瓙𦁚𱋊𨮉笥𢵎啇𭶛𨊼鼻𡦆丝𤿖珳𭴆𪍆琶𭕈辋𪦋珃𣐗 
汈𫋇𤦣㴙𣰳𮑱𰵞𤩿𥎮呭𨤕誷𧈣𪳟𤶼𭸫𩌛䵜𠄲𡕏𧤫𭐅𢒛𭲀𰅜灺𠇖瓃揯𤞩𡣸𡕂𭨈𢬯𮍇𦚼㚗坂㺔𮍠㻤𠈑𰹪𠥳𠐖哌𤻯𬯣𧜑𬎎𣆸𣋔𤤚𬢊𩂬𦰆�𢃆𫭹藑藑𧝹𱪈電𩤰鯎䏜𢂩𪁦𫥆𢟐𧌠𩦳釯𨠕坥本𥤃𦥕姒𤇚𠭯 
𡩍𣎄㮓𫖼𡛽鯸脈𣄺𰄴𤥖𱼙𡒸𮥮𦗱𫃾𧇘䞧赬㻇𡖞玧𢅏𣌶𠶃𫔞𦲳𢹈𰯧𤟺𤫍𭹐𠾗髀𥈊㰀犮𱻅燫𪼜𢆵𭹂𬝍颦比𠓓𤓫珚𧤉𫀺𪝰𭛮焯氵𦟁𤖄𰘄狰𣀊�傮傮𬈉𱧄臼𦷵𤕱槜垱秣𪢜𡗄晄𡥸𦁏𨌬琤𬖜𲎭䗔𨲌𥙎𤨞 
𦚯𫦵𮟬䨤羊𡧥𤫥𱲑𭿩𭹭磃拽𠹴𧚜𤫁𪣍𠰹𭹵𭳃𱸉𬜎搗𭹠𣢅咅𥿯𫆑𱗓䧦珨𠙚𬨚𠘊𮏪刀䪆鸟𧙸𢄂𧟺𪙭𨹃𠻷𥝢𩧇䙈𤧏𫫓𬇠𰼾𣃶𤫚𭓥㰎𤆈萆謙𨶍𫧴𨽄𪜊段砠𥦩㷮𩶴𡟈𡣍𤐬𩈘𱑑𩺹𡞤㻼𤨜𠏻璃䔃𤥃𩦝𠵰𤩶
甉𪞞𣺣𰚓級𰳈襘𠓚也𮐘𰬾睝𦣩𫩺䉕𪥻𮗗𩬻𢔨𧉮𤧐𡯦𩊈𫧓玶𬺶㻄𠑆𥪖𪼌𤨉𱯤𪒟𪼅巯𬜮眨蚰䋸𪣜𠒬𡶲𮇊𫕯訴𪱯𣷋𠆦𲁥𧴚毈𡳟瑈𥻼㬝甑𤛢玷㪕�尢尢𮭙岋𦘢𣵪亼𡑱瑮陁𡡑𡭐𱲐嬴𢂠𢽭𬜆𮠗晡玣琲戛 
兀𬝐埩巾𱈇𣻐巠𱜳𱛽迳辵𢫕𠂢𠃏𬙀𮦛𮐥瑐𰡈𰺾𨺑𫼜瓌𬼋𡍎𨷧𰇝𣽗琝𣁬𰶕䊨𪁭斷𲊗镰𱝯𨄊斖𤭜琌𤩧𢇍𮋭𬜧𧋅鲺𭹘𨪥𥠻𢍤𣫋𥀛𤨌𫃨𤧷㻌𠨑𡭼𡁇𦞴𪲇𮅢𨖯𥰽𮄁阜𨞠嬝𩺐𤫠詞𦱥𡬭𣽊䔘玐玽𤩜𡘞𪆯朇
𦂰𪡦𱊓𰱵𰵌𥚁𤪙𭕚𧈕挡𦇽𩳁𭆬𧾩𤤞禬𱸓𫫆𰯭閼䠪𭬹𭚪𭌢𦚖𱎞石㜑𥌀𦓟䁦㕽𪜲舨𦠇𣜔𱫰𪞀𰎯𰤕玞局𠢟𧤢𢾨𫳚𨑑𬗫𥙩𦣞𪯦𠅼𨝋𢍖𱙋眜𩐈𡗭𣯑瑷𱧈爼㣌𱄸丙鵫𡡛𣳍𧙊𤤽𡄫𫏆扡𦑷體𮢻三詠𤥶𱰣𭔳贬
膥遞𬷱山𣓐䃣䂉𮍡𧒓𠮨𤦮𱯔𮧰𡻂𦭩𦿘𤪢𡦵𪻣𦁰䎖玹𭅷𮙴丵𩬣𠠻𤴘𭵵𤟣𭲄憹絁珿𰸱襾㣸𪕻𤦔萋𧕣𢂪𱎟㪑𡿘糸𰇆𠅦𡹠𡹵瘀𭺫𩺊庵𤪹𩭟琄𦱉掩㻻�苫苫珹𫧯𡯸䥥𮊅𰚊𭣞𭭞𢄓𭻵𲂢㼇𤩽淹𣌇𱄹㑺㮚𤍙 
𫲥赪𦼬璦𤥉𠲨𭷵蛐杨𡶉𠺯𩟊𧕳腶𥠗𬀼𭨟咿𢆟𡺡䃀𭀨𥏰𰮗𭉍𥯈𥣎𩩵珜鴔𭊜𭪋𮡾夆𤧹𪱮㞚𦱣𬀭熓𫵑𭉣𡝜㥈𰡹𫼓𠚜掉𣉥𠎖𬉻𤧢𨩙𬎊𣻏𦑤𭆰𥙌霢璕𦏞𡲿鏳𫽾𭬴𧉝𬨌𫂯𬃇𣵈𭹋𭽤𧳝𨗱𬟸𦹫𰊓㼹璚䇑灔䵄
𨵍钖𮙗𣶺㺳𧰳臁𤧾𤼜𭼃𱥶㠷𩧎𩊄𫜔𥠱𡜵𡊡𪈉𥶰𮖼𬉩𱈨耒𪹔䤪𠌽𪼬𤩵𭡐𤦵寡𥟮𦰈筁𰀺𪾗䄁𮊤㐌𡠘乶㻎𮢋瀞𥍿𡌧㝃𫡟
[I 230419 20:24:41 He2Zi:60]  jizi:591 -> 537 useless k:54 ㇕㇚㇉㇦㇋㇜㇧㇙㇒㇪㇤㇞⿽㇨⿾㇄㇁㇮㇘㇎⿼⿿㇟㇡㇑㇭㇛㇠㇬㇣㇂㇢㇥㇐㇊㇩㇈㇆㇌㇇㇫㇝㇔㇯㇗㇅㇍㇃㇓㒳㇏㇀
㇖爫
[I 230419 20:24:41 He2Zi:92] HeZi build success -> demo/He2Ji.txt  demo/JiZi.txt

"""



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from os import path
import os

# packages = find_packages(exclude=['test'],include=['ZiTokenizer'])
packages = find_packages(exclude=['test'])
print(packages)

here = os.path.abspath(os.path.dirname(__file__))

with open(path.join(here, 'readme.md')) as f:
    long_description = f.read()


setup(
    name="ZiTokenizer",
    packages=find_packages(),
    # py_modules = ['ZiTokenizer.UnicodeTokenizer', 'ZiTokenizer.ZiCutter', 'ZiTokenizer.ZiTokenizer', 'ZiTokenizer.ZiSegmenter'],
    version='0.0.8',
    description='ZiTokenizer: tokenize world text as Zi',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.0',
    install_requires=[
        "logzero",
        # "UnicodeTokenizer",
        "pyahocorasick",
        # "ZiCutter"
    ],

    # include_package_data=True,
    package_data={
        # 引入任何包下面的 *.txt、*.rst 文件
        "": ["*.txt", "*.rst"],
        # 引入 hello 包下面的 *.msg 文件
        "ZiTokenizer": ["*.txt"],
        "ZiTokenizer": ["languages/*/*.txt"],
    },
    # data_files=[('data', ['data/ChaiZi.txt'])],

    url='https://github.com/laohur/ZiCutter',
    keywords=['UnicodeTokenizer', 'ZiCutter', 'ZiTokenizer',
              'Tokenizer', 'Unicode', 'laohur'],
    author='laohur',
    author_email='laohur@gmail.com',
    license='[Anti-996 License](https: // github.com/996icu/996.ICU/blob/master/LICENSE)',
)

"""
python setup.py sdist
python setup.py bdist_wheel
twine upload dist/*
"""




import collections
import os
import random
import math

from logzero import logger

from .UnicodeTokenizer import UnicodeTokenizer
from .ZiCutter import ZiCutter
from .ZiSegmenter import ZiSegmenter


class ZiTokenizer:
    def __init__(self, dir=None, do_lower_case=True, max_split=3, unk_k=10, split_digit=False, never_split=[]) -> None:
        self.do_lower_case = do_lower_case
        self.max_split = max_split
        self.split_digit = split_digit
        self.dir = dir
        self.never_split = set(x for x in never_split)
        self.token2index = collections.OrderedDict()
        self.UNKs = [f"##{x}" for x in range(unk_k)]  # 10
        if dir == None:
            dir = "languages/global"
        if dir:
            self.load(dir)

    def load(self, folder):
        vocab = []
        root_words = []
        prefixs = []
        suffixs = []
        vocab_path = os.path.join(folder, "vocab.txt")
        assert os.path.exists(vocab_path)
        vocab = open(vocab_path).read().splitlines()
        for i, x in enumerate(vocab):
            if len(x) > 1:
                if x[:2] == '--':
                    suffixs.append(x[2:])
                    continue
                if x[-2:] == '--':
                    prefixs.append(x[:-2])
                    continue
            root_words.append(x)
        logger.info(
            f" {vocab_path} load vocab:{len(vocab)} root:{len(root_words)} prefix:{len(prefixs)} suffix:{len(suffixs)} ")

        self.vocab = vocab
        for i, x in enumerate(vocab):
            self.token2index[x] = i

        root_words = set(root_words)
        never_split = self.never_split | root_words
        self.unicodeTokenizer = UnicodeTokenizer(do_lower_case=self.do_lower_case, never_split=never_split, split_digit=self.split_digit)
        self.ziCutter = ZiCutter(folder)
        self.ziSegmenter = ZiSegmenter(
            root_words=root_words, prefixs=prefixs, suffixs=suffixs, max_split=self.max_split)

    def cutRare(self, token) -> str:
        point = sum(ord(x) for x in token) % 10
        return f"##{point}"

    def token_word(self, word):
        [heads, root, tails] = self.ziSegmenter.token_word(word)
        if root:
            for i in range(len(heads)):
                heads[i] += '--'
            for i in range(len(tails)):
                tails[i] = '--' + tails[i]
            tokens = heads+[root]+tails
            return tokens
        ids = self.ziCutter.cutHanzi(word)
        if ids == word:
            tokens = [self.cutRare(word)]
        else:
            tokens = list(ids)
        return tokens

    def build(self, word_freq, total, folder, min_ratio=1.5e-6, min_freq=2, vocab_size=-1):
        bottom = max(min_freq, (total*min_ratio))
        bottom_word = bottom
        logger.info(
            f"min_ratio:{min_ratio} min_freq:{min_freq} bottom:{bottom:.2f} vocab_size:{vocab_size}")
        hot = set()
        for k, v in word_freq:
            if v >= bottom:
                hot.add(k)
                bottom_word = v
            else:
                break
            if 0 < vocab_size < len(hot):
                break
        bottom_char = max(min_freq, bottom_word/4)
        logger.info(f"  bottom_word:{bottom_word:.2f} bottom_char:{bottom_char:.2f}")
        chars = [k for k, v in word_freq if len(k) == 1 and v >= bottom_char and k not in hot]
        root_words = self.never_split | hot | set(chars)
        logger.info(
            f" words:{len(word_freq)} hot:{len(hot)} chars:{len(chars)} root_words:{len(root_words)}")

        ziCutter = ZiCutter(folder)
        ziCutter.build(folder, root_words)
        root_words |= set(ziCutter.vocab)
        ziSegmenter = ZiSegmenter(root_words=root_words)

        logger.info("  === token_root ===  ")
        sample = random.choices(word_freq, k=5)
        for k, v in sample:
            [prefix, root, suffix] = ziSegmenter.token_root(k)
            row = [k, v, prefix, root, suffix]
            logger.info((row))

        prefix_counter = collections.Counter()
        suffix_counter = collections.Counter()
        # ignore rare words and decline bottom may save time
        for k, v in word_freq:
            if k in root_words:
                continue
            [prefix, root, suffix] = ziSegmenter.token_root(k)
            if not root:
                continue
            if prefix:
                prefix_counter[prefix] += v
            if suffix:
                suffix_counter[suffix] += v
        del word_freq
        prefixs = [k for k, v in prefix_counter.items() if v >= bottom_word]
        del prefix_counter
        suffixs = [k for k, v in suffix_counter.items() if v >= bottom_word]
        del suffix_counter
        logger.info(
            f"root_words:{len(root_words)} prefixs:{len(prefixs)} suffixs:{len(suffixs)}")

        prefixs = [x+'--' for x in prefixs]
        root_words = [x for x in root_words]
        suffixs = ['--'+x for x in suffixs]
        vocab = self.UNKs + sorted(root_words)+sorted(prefixs)+sorted(suffixs)
        vocab_path = os.path.join(folder, "vocab.txt")
        with open(vocab_path, 'w') as f:
            for x in vocab:
                f.write(x+'\n')
        logger.info(f"save  vocab { len(vocab) }  -->{vocab_path} ")
        self.load(folder)

    def tokenize(self, line):
        words = self.unicodeTokenizer.tokenize(line)
        tokens = []
        for word in words:
            if not word:
                continue
            if word in self.token2index:
                tokens.append(word)
            else:
                cuts = self.token_word(word)
                tokens += cuts
        tokens = [x for x in tokens if x]
        return tokens

    def tokens2indexs(self, tokens):
        idxs = [self.token2index[x] for x in tokens]
        return idxs

    def indexs2tokens(self, indexs):
        indexs = [self.vocab[x] for x in indexs]
        return indexs

    def encode(self, line):
        tokens = self.tokenize(line)
        indexs = self.tokens2indexs(tokens)
        return indexs

    def tokens2words(self, tokens):
        ts = tokens[:1]
        for i in range(1, len(tokens)):
            x = tokens[i]
            if ts[-1].endswith('--'):  # prefix
                ts[-1] = ts[-1][:-2]+x
            elif x.startswith('--'):  # suffix
                ts[-1] += x[2:]
            else:
                ts.append(x)
        return ts

    def decode(self, indexs):
        tokens = self.indexs2tokens(indexs)
        ts = self.tokens2words(tokens)
        words = self.ziCutter.combineHanzi(ts)
        return words

    def tokens2line(self, tokens):
        line = self.unicodeTokenizer.detokenize(tokens)
        return line
